#include "common.h"
#include "random.h"
#include <stdlib.h>
#include <string.h>
extern FILE *logfile;
EuObject::EuObject()
{
   type = INTEGER;
   iVal = 0;
   size = 0;
}
EuObject::EuObject(EuObject *obj)
{
   type = INTEGER;
   size = 0;
   actualSize = 0;
   iVal = 0;
   mov(obj);
}
EuObject::~EuObject()
{
   if (type == SEQUENCE)
   {
      if (children)
            delete [] children;
   }
}
void EuObject::fastmov(EuObject *obj)
{
   type = obj->type;
   if (type == INTEGER)
   {
      iVal = obj->iVal;
   } else if (type == ATOM)
   {
      aVal = obj->aVal;
   } else {
      children = obj->children;
      size = obj->size;
      actualSize = obj->actualSize;
   }
}
void EuObject::mov(EuObject *obj)
{
   if (obj->type == INTEGER)
   {
      if (type == SEQUENCE)
      {
         if (children)
            delete [] children;
      }
      type = INTEGER;
      iVal = obj->iVal;
   } else if (obj->type == ATOM)
   {
      if (type == SEQUENCE)
      {
         if (children)
            delete [] children;
      }
      type = ATOM;
      aVal = obj->aVal;
   } else {
      if (type == SEQUENCE) {
            size = obj->size;
            if (size > actualSize)
            {
               EuObject *newElements = (EuObject *)realloc(children,sizeof(EuObject)*size);
               actualSize = size;
               if (newElements != children)
                  delete [] children;
               children = newElements;
            }
      } else {
            type = SEQUENCE;
            size = obj->size;
            actualSize = size;
            children = new EuObject[size];
      }
      for (long i = 0; i < size; i++)
          children[i].mov(&obj->children[i]);
   }
}
void EuObject::movdi(EuObject *index, EuObject *val)
{
   children[index->iVal-1].mov(val);
}
void EuObject::movsi(EuObject *index, EuObject *val)
{
   mov(&val->children[index->iVal-1]);
}
void EuObject::append(EuObject *obj)
{
   if (type == SEQUENCE)
   {
      size++;
      if (size > actualSize)
      {
         EuObject *newElements = (EuObject *)realloc(children,sizeof(EuObject)*size);
         if (newElements == 0)
         {
            printf("newElements == 0 %d %d\n",size,actualSize);
            abort();
         }
         actualSize = size;
         if (newElements != children)
            delete [] children;
         children = newElements;
      }
      children[size-1].mov(obj);
   } else {
      EuObject *c = new EuObject[2];
      c[0].mov(this);
      c[1].mov(obj);
      children = c;
      size = 2;
      actualSize = 2;
      type = SEQUENCE;
   }
}
void EuObject::prepend(EuObject *obj)
{
   if (type == SEQUENCE)
   {
      size++;
      if (size > actualSize)
      {
         EuObject *newElements = (EuObject *)realloc(children,sizeof(EuObject)*size);
         actualSize = size;
         if (newElements != children)
            delete [] children;
         children = newElements;
      }
      for (long i = size-2; i >= 0; i--)
          children[i+1].fastmov(&children[i]);
      children[0].mov(obj);
   } else {
      EuObject *c = new EuObject[2];
      c[0].mov(obj);
      c[1].mov(this);
      children = c;
      size = 2;
      actualSize = 2;
      type = SEQUENCE;
   }
}
void EuObject::movlen(EuObject *obj)
{
   if (type == SEQUENCE)
   {
      if (children)
         delete [] children;
   }
   type = INTEGER;
   iVal = obj->size;
}
void EuObject::repeat(EuObject *n)
{
   EuObject *objs = new EuObject[n->iVal];
   for (long i = 0; i < n->iVal; i++)
       objs[i].mov(this);
   if (type == SEQUENCE)
   {
      if (children)
         delete [] children;
   }
   children = objs;
   type = SEQUENCE;
   size = n->iVal;
   actualSize = size;
}
void EuObject::not()
{
	if (type == INTEGER)
	{
      iVal = (!iVal);
	} else if (type == ATOM) {
      type = INTEGER;
      iVal = (!((long)aVal));
	} else {
      for (long i = 0; i < size; i++)
          children[i].not();
	}
}
void EuObject::and(EuObject *obj)
{
	if (type == INTEGER)
	{
		if (obj->type == INTEGER)
			iVal = iVal && obj->iVal;
		else if (obj->type == ATOM) {
         if ((iVal == 0) || (obj->aVal == 0))
         {
            iVal = 0;
         } else {
            iVal = 1;
         }
		} else {
			printf("Can't and integer, sequence\n");
			abort();
		}
	} else if (type == ATOM) {
		if (obj->type == INTEGER) {
         type = INTEGER;
         if ((aVal == 0) || (obj->iVal == 0))
         {
            iVal = 0;
         } else {
            iVal = 1;
         }
      }
		else if (obj->type == ATOM) {
         type = INTEGER;
         if ((aVal == 0) || (obj->aVal == 0))
         {
            iVal = 0;
         } else {
            iVal = 1;
         }
		} else {
			printf("Can't and atom, sequence\n");
			abort();
		}
	} else {
		if (obj->type != SEQUENCE) {
         for (long i = 0; i < size; i++)
             children[i].and(obj);
		} else {
			if (size != obj->size) {
            for (long i = 0; i < size; i++)
                children[i].and(obj);
			} else {
            for (long i = 0; i < size; i++)
                children[i].and(&obj->children[i]);
			}
		}
	}
}
void EuObject::or(EuObject *obj)
{
	if (type == INTEGER)
	{
		if (obj->type == INTEGER)
			iVal = iVal || obj->iVal;
		else if (obj->type == ATOM) {
         if ((iVal == 0) && (obj->aVal == 0))
         {
            iVal = 0;
         } else {
            iVal = 1;
         }
		} else {
			printf("Can't or integer, sequence\n");
			abort();
		}
	} else if (type == ATOM) {
		if (obj->type == INTEGER) {
         type = INTEGER;
         if ((aVal == 0) && (obj->iVal == 0))
         {
            iVal = 0;
         } else {
            iVal = 1;
         }
      }
		else if (obj->type == ATOM) {
         type = INTEGER;
         if ((aVal == 0) && (obj->aVal == 0))
         {
            iVal = 0;
         } else {
            iVal = 1;
         }
		} else {
			printf("Can't or atom, sequence\n");
			abort();
		}
	} else {
		if (obj->type != SEQUENCE) {
         for (long i = 0; i < size; i++)
             children[i].or(obj);
		} else {
			if (size != obj->size) {
            for (long i = 0; i < size; i++)
                children[i].or(obj);
			} else {
            for (long i = 0; i < size; i++)
                children[i].or(&obj->children[i]);
			}
		}
	}
}
void EuObject::xor(EuObject *obj)
{
	if (type == INTEGER)
	{
		if (obj->type == INTEGER)
			iVal = (iVal ^ obj->iVal) != 0;
		else if (obj->type == ATOM) {
         if ((iVal != 0) ^ (obj->aVal != 0))
         {
            iVal = 1;
         } else {
            iVal = 0;
         }
		} else {
			printf("Can't xor integer, sequence\n");
			abort();
		}
	} else if (type == ATOM) {
		if (obj->type == INTEGER) {
         type = INTEGER;
         if ((aVal != 0) ^ (obj->iVal != 0))
         {
            iVal = 1;
         } else {
            iVal = 0;
         }
      } else if (obj->type == ATOM) {
         type = INTEGER;
         if ((aVal != 0) ^ (obj->aVal != 0))
         {
            iVal = 1;
         } else {
            iVal = 0;
         }
		} else {
			printf("Can't xor atom, sequence\n");
			abort();
		}
	} else {
		if (obj->type != SEQUENCE) {
         for (long i = 0; i < size; i++)
             children[i].xor(obj);
		} else {
			if (size != obj->size) {
            for (long i = 0; i < size; i++)
                children[i].xor(obj);
			} else {
            for (long i = 0; i < size; i++)
                children[i].xor(&obj->children[i]);
			}
		}
	}
}
void EuObject::concen(EuObject *obj)
{
   if (obj->type == SEQUENCE)
   {
      long start;
      if (type == SEQUENCE)
      {
         start = size;
         size += obj->size;
         if (size > actualSize)
         {
            EuObject *newElements = (EuObject *)realloc(children,sizeof(EuObject)*size);
            actualSize = size;
            if (newElements != children)
               delete [] children;
            children = newElements;
         }
      } else {
         actualSize = size;
         children = new EuObject[size];
         start = 0;
      }
      long n=0;
      for (long i = start; i < size; i++)
      {
          children[i].mov(&obj->children[n]);
          n++;
      }
   } else {
     if (type == SEQUENCE)
     {
         size++;
         if (size > actualSize)
         {
            EuObject *newElements = (EuObject *)realloc(children,sizeof(EuObject)*size);
            actualSize = size;
            if (newElements != children)
               delete [] children;
            children = newElements;
         }
         children[size-1].mov(obj);
     } else {
         EuObject *c = new EuObject[2];
         c[0].mov(this);
         c[1].mov(obj);
         children = c;
         size = 2;
         actualSize = 2;
         type = SEQUENCE;
     }
   }
}
void EuObject::isinteger(EuObject *obj)
{
   if (type == SEQUENCE)
   {
      if (children)
         delete [] children;
   }
   type = INTEGER;
   iVal = 0;
   if (obj->type == INTEGER)
      iVal = 1;
   else if (obj->type == ATOM)
   {
      if (floor(obj->aVal) == obj->aVal)
      {
         iVal = 1;
      }
   }
}
void EuObject::isatom(EuObject *obj)
{
   if (type == SEQUENCE)
   {
      if (children)
         delete [] children;
   }
   type = INTEGER;
   iVal = (obj->type != SEQUENCE);
}
void EuObject::issequence(EuObject *obj)
{
   if (type == SEQUENCE)
   {
      if (children)
         delete [] children;
   }
   type = INTEGER;
   iVal = (obj->type == SEQUENCE);
}
void EuObject::andbits(EuObject *obj)
{
	if (type == INTEGER)
	{
		if (obj->type == INTEGER)
			iVal = iVal & obj->iVal;
		else if (obj->type == ATOM) {
         iVal = iVal & (long)obj->aVal;
		} else {
			printf("Can't andbits integer, sequence\n");
			abort();
		}
	} else if (type == ATOM) {
		if (obj->type == INTEGER) {
         type = INTEGER;
         iVal = (long)aVal & obj->iVal;
      }
		else if (obj->type == ATOM) {
         type = INTEGER;
         iVal = (long)aVal & (long)obj->aVal;
		} else {
			printf("Can't andbits atom, sequence\n");
			abort();
		}
	} else {
		if (obj->type != SEQUENCE) {
         for (long i = 0; i < size; i++)
             children[i].andbits(obj);
		} else {
			if (size != obj->size) {
            for (long i = 0; i < size; i++)
                children[i].andbits(obj);
			} else {
            for (long i = 0; i < size; i++)
                children[i].andbits(&obj->children[i]);
			}
		}
	}
}
void EuObject::orbits(EuObject *obj)
{
	if (type == INTEGER)
	{
		if (obj->type == INTEGER)
			iVal = iVal | obj->iVal;
		else if (obj->type == ATOM) {
         iVal = iVal | (long)obj->aVal;
		} else {
			printf("Can't orbits integer, sequence\n");
			abort();
		}
	} else if (type == ATOM) {
		if (obj->type == INTEGER) {
         type = INTEGER;
         iVal = (long)aVal | obj->iVal;
      }
		else if (obj->type == ATOM) {
         type = INTEGER;
         iVal = (long)aVal | (long)obj->aVal;
		} else {
			printf("Can't orbits atom, sequence\n");
			abort();
		}
	} else {
		if (obj->type != SEQUENCE) {
         for (long i = 0; i < size; i++)
             children[i].orbits(obj);
		} else {
			if (size != obj->size) {
            for (long i = 0; i < size; i++)
                children[i].orbits(obj);
			} else {
            for (long i = 0; i < size; i++)
                children[i].orbits(&obj->children[i]);
			}
		}
	}
}
void EuObject::xorbits(EuObject *obj)
{
	if (type == INTEGER)
	{
		if (obj->type == INTEGER)
			iVal = iVal ^ obj->iVal;
		else if (obj->type == ATOM) {
           iVal = iVal ^ (long)obj->aVal;
		} else {
			printf("Can't xorbits integer, sequence\n");
			abort();
		}
	} else if (type == ATOM) {
		if (obj->type == INTEGER) {
         type = INTEGER;
         iVal = (long)aVal ^ obj->iVal;
      } else if (obj->type == ATOM) {
         type = INTEGER;
         iVal = (long)aVal ^ (long)obj->aVal;
		} else {
			printf("Can't xorbits atom, sequence\n");
			abort();
		}
	} else {
		if (obj->type != SEQUENCE) {
         for (long i = 0; i < size; i++)
             children[i].xorbits(obj);
		} else {
			if (size != obj->size) {
            for (long i = 0; i < size; i++)
                children[i].xorbits(obj);
			} else {
            for (long i = 0; i < size; i++)
                children[i].xorbits(&obj->children[i]);
			}
		}
	}
}
void EuObject::Floor()
{
   if (type == ATOM)
   {
      type = INTEGER;
      iVal = (long)floor(aVal);
   } else if (type == SEQUENCE)
   {
      for (long i = 0; i < size; i++)
          children[i].Floor();
   }
}
void EuObject::Sqrt()
{
   if (type == ATOM)
   {
      aVal = sqrt(aVal);
   } else if (type == INTEGER) {
      type = ATOM;
      aVal = sqrt(iVal);
   } else if (type == SEQUENCE)
   {
      for (long i = 0; i < size; i++)
          children[i].Sqrt();
   }
}
void EuObject::Cos()
{
   if (type == ATOM)
   {
      aVal = cos(aVal);
   } else if (type == INTEGER) {
      type = ATOM;
      aVal = cos(iVal);
   } else if (type == SEQUENCE)
   {
      for (long i = 0; i < size; i++)
          children[i].Cos();
   }
}
void EuObject::arccos()
{
   if (type == ATOM)
   {
      aVal = acos(aVal);
   } else if (type == INTEGER) {
      type = ATOM;
      aVal = acos(iVal);
   } else if (type == SEQUENCE)
   {
      for (long i = 0; i < size; i++)
          children[i].arccos();
   }
}
void EuObject::Sin()
{
   if (type == ATOM)
   {
      aVal = sin(aVal);
   } else if (type == INTEGER) {
      type = ATOM;
      aVal = sin(iVal);
   } else if (type == SEQUENCE)
   {
      for (long i = 0; i < size; i++)
          children[i].Sin();
   }
}
void EuObject::arcsin()
{
   if (type == ATOM)
   {
      aVal = asin(aVal);
   } else if (type == INTEGER) {
      type = ATOM;
      aVal = asin(iVal);
   } else if (type == SEQUENCE)
   {
      for (long i = 0; i < size; i++)
          children[i].arcsin();
   }
}
void EuObject::Tan()
{
   if (type == ATOM)
   {
      aVal = tan(aVal);
   } else if (type == INTEGER) {
      type = ATOM;
      aVal = tan(iVal);
   } else if (type == SEQUENCE)
   {
      for (long i = 0; i < size; i++)
          children[i].Tan();
   }
}
void EuObject::arctan()
{
   if (type == ATOM)
   {
      aVal = atan(aVal);
   } else if (type == INTEGER) {
      type = ATOM;
      aVal = atan(iVal);
   } else if (type == SEQUENCE)
   {
      for (long i = 0; i < size; i++)
          children[i].arctan();
   }
}
void EuObject::Log()
{
   if (type == ATOM)
   {
      aVal = log10(aVal);
   } else if (type == INTEGER) {
      type = ATOM;
      aVal = log10(iVal);
   } else if (type == SEQUENCE)
   {
      for (long i = 0; i < size; i++)
          children[i].Log();
   }
}
double myRemainder(double a, double b)
{
   bool neg = false;
   double v1, v2;
   if (b < 0)
      b = -b;
   if (a < 0)
   {
      neg = true;
      a = -a;
   }
   if (b == 0)
   {
      printf("Can't remainder x, 0");
      abort();
   }
   v1 = 0;
   while (1)
   {
      v2 = v1 + b;
      if (v2 >= a)
         break;
      v1 = v2;
   }
   if (neg)
   {
      return -(a-v1);
   } else {
      return a-v1;
   }
}
void EuObject::remainder(EuObject *obj)
{
	if (type == INTEGER)
	{
		if (obj->type == INTEGER)
			iVal = iVal % obj->iVal;
		else if (obj->type == ATOM) {
         type = ATOM;
         aVal = myRemainder((double)iVal,obj->aVal);
		} else {
			printf("Can't remainder integer, sequence\n");
			abort();
		}
	} else if (type == ATOM) {
		if (obj->type == INTEGER) {
         type = INTEGER;
         aVal = myRemainder((double)iVal,obj->aVal);
      }
		else if (obj->type == ATOM) {
         aVal = myRemainder((double)iVal,obj->aVal);
		} else {
			printf("Can't remainder atom, sequence\n");
			abort();
		}
	} else {
		if (obj->type != SEQUENCE) {
         for (long i = 0; i < size; i++)
             children[i].remainder(obj);
		} else {
			if (size != obj->size) {
            for (long i = 0; i < size; i++)
                children[i].remainder(obj);
			} else {
            for (long i = 0; i < size; i++)
                children[i].remainder(&obj->children[i]);
			}
		}
	}
}
void EuObject::rand(EuObject *v)
{
   if (v->type == INTEGER)
   {
      if (type == SEQUENCE)
      {
         if (children)
            delete [] children;
      }
      type = INTEGER;
      iVal = (long)floor(myrandom()*v->iVal)+1;
   } else if (v->type == ATOM)
   {
      if (type == SEQUENCE)
      {
         if (children)
            delete [] children;
      }
      type = INTEGER;
      iVal = (long)floor(myrandom()*floor(v->iVal))+1;
   } else {
      if (type == SEQUENCE)
      {
         size = v->size;
         if (size > actualSize)
         {
            EuObject *newElements = (EuObject *)realloc(children,sizeof(EuObject)*size);
            actualSize = size;
            if (newElements != children)
               delete [] children;
            children = newElements;
         }
      } else {
         size = v->size;
         type = SEQUENCE;
         children = new EuObject[size];
      }
      for (long i = 0; i < size; i++)
          children[i].rand(&v->children[i]);
   }
}
void EuObject::power(EuObject *obj)
{
	if (type == INTEGER)
	{
		if (obj->type == INTEGER)
			iVal = (long)pow((double)iVal,(double)obj->iVal);
		else if (obj->type == ATOM) {
         type = ATOM;
         aVal = pow((double)iVal,obj->aVal);
		} else {
			printf("Can't power integer, sequence\n");
			abort();
		}
	} else if (type == ATOM) {
		if (obj->type == INTEGER) {
         aVal = pow(aVal,(double)obj->iVal);
		} else if (obj->type == ATOM) {
         aVal = pow(aVal,obj->aVal);
		} else {
			printf("Can't power atom, sequence\n");
			abort();
		}
	} else {
		if (obj->type != SEQUENCE) {
         for (long i = 0; i < size; i++)
             children[i].power(obj);
		} else {
			if (size != obj->size) {
            for (long i = 0; i < size; i++)
                children[i].power(obj);
			} else {
            for (long i = 0; i < size; i++)
                children[i].power(&obj->children[i]);
			}
		}
	}
 }
void EuObject::ce(EuObject *obj)
{
   if (type == INTEGER)
   {
      if (obj->type == INTEGER)
      {
         setRv(iVal == obj->iVal);
      } else if (obj->type == ATOM)
      {
         setRv((double)iVal == obj->aVal);
      } else {
        printf("can't ce integer, sequence.\n");
        abort();
      }
   } else if (type == ATOM)
   {
      if (obj->type == ATOM)
      {
         setRv(aVal == obj->aVal);
      } else if (obj->type == INTEGER)
      {
         setRv(aVal == (double)obj->iVal);
      } else {
        printf("can't ce atom, sequence.\n");
        abort();
      }
   } else {
     printf("can't ce sequences.\n");
   }
}
void EuObject::cle(EuObject *obj)
{
   if (type == INTEGER)
   {
      if (obj->type == INTEGER)
      {
         setRv(iVal <= obj->iVal);
      } else if (obj->type == ATOM)
      {
         setRv((double)iVal <= obj->aVal);
      } else {
        printf("can't cle integer, sequence.\n");
        abort();
      }
   } else if (type == ATOM)
   {
      if (obj->type == ATOM)
      {
         setRv(aVal <= obj->aVal);
      } else if (obj->type == INTEGER)
      {
         setRv(aVal <= (double)obj->iVal);
      } else {
        printf("can't cle atom, sequence.\n");
        abort();
      }
   } else {
     printf("can't cle sequences.\n");
   }
}
void EuObject::cge(EuObject *obj)
{
   if (type == INTEGER)
   {
      if (obj->type == INTEGER)
      {
         setRv(iVal >= obj->iVal);
      } else if (obj->type == ATOM)
      {
         setRv((double)iVal >= obj->aVal);
      } else {
        printf("can't cge integer, sequence.\n");
        abort();
      }
   } else if (type == ATOM)
   {
      if (obj->type == ATOM)
      {
         setRv(aVal >= obj->aVal);
      } else if (obj->type == INTEGER)
      {
         setRv(aVal >= (double)obj->iVal);
      } else {
        printf("can't cge atom, sequence.\n");
        abort();
      }
   } else {
     printf("can't cge sequences.\n");
   }
}
void EuObject::cl(EuObject *obj)
{
   if (type == INTEGER)
   {
      if (obj->type == INTEGER)
      {
         setRv(iVal < obj->iVal);
      } else if (obj->type == ATOM)
      {
         setRv((double)iVal < obj->aVal);
      } else {
        printf("can't cl integer, sequence.\n");
        abort();
      }
   } else if (type == ATOM)
   {
      if (obj->type == ATOM)
      {
         setRv(aVal < obj->aVal);
      } else if (obj->type == INTEGER)
      {
         setRv(aVal < (double)obj->iVal);
      } else {
        printf("can't cl atom, sequence.\n");
        abort();
      }
   } else {
     printf("can't cl sequences.\n");
   }
}
void EuObject::cg(EuObject *obj)
{
   if (type == INTEGER)
   {
      if (obj->type == INTEGER)
      {
         setRv(iVal > obj->iVal);
      } else if (obj->type == ATOM)
      {
         setRv((double)iVal > obj->aVal);
      } else {
        printf("can't cg integer, sequence.\n");
        abort();
      }
   } else if (type == ATOM)
   {
      if (obj->type == ATOM)
      {
         setRv(aVal > obj->aVal);
      } else if (obj->type == INTEGER)
      {
         setRv(aVal > (double)obj->iVal);
      } else {
        printf("can't cg atom, sequence.\n");
        abort();
      }
   } else {
     printf("can't cg sequences.\n");
   }
}
void EuObject::cmp(EuObject *obj)
{
   EuObject *a;
   EuObject *b;
   if (type == INTEGER)
   {
      if (obj->type == INTEGER)
      {
         if (iVal == obj->iVal)
            setRv(0);
         else if (iVal < obj->iVal)
            setRv(-1);
         else
            setRv(1);
      } else if (obj->type == ATOM)
      {
         if ((double)iVal == obj->aVal)
            setRv(0);
         else if ((double)iVal < obj->aVal)
            setRv(-1);
         else
            setRv(1);
      } else {
        setRv(-1);
      }
   } else if (type == ATOM)
   {
      if (obj->type == ATOM)
      {
         if (aVal == obj->aVal)
            setRv(0);
         else if (aVal < obj->aVal)
            setRv(-1);
         else
            setRv(1);
      } else if (obj->type == INTEGER)
      {
         if (aVal == (double)obj->iVal)
            setRv(0);
         else if (aVal < (double)obj->iVal)
            setRv(-1);
         else
            setRv(1);
      } else {
         setRv(-1);
      }
   } else {
     if (obj->type != SEQUENCE)
     {
       setRv(1);
     } else {
       for (long i = 0; i < size; i++) {
           children[i].cmp(&obj->children[i]);
           if (rv->iVal != 0)
              return;
       }
       if (a == 0)
       {
         if (b == 0)
         {
            setRv(0);
         } else {
            setRv(1);
         }
       } else {
         setRv(-1);
       }
     }
   }
}
void EuObject::equ(EuObject *obj)
{
   cmp(obj);
   if (rv->iVal != 0)
      rv->iVal = 1;
}
